@@ -2,13 +2,16 @@
 Module for AnnotationSet class which represents a named collection of annotations which can arbitrarily overlap.
 """
 
-from typing import Any, List, Tuple, Union, Dict, Set, KeysView, Iterator, Generator
+# TODO: when should two sets be equal? Currently object identity is requried!
+
+from typing import Any, List, Union, Dict, Set, KeysView, Iterator, Generator
+from collections.abc import Iterable
 from collections import defaultdict
 import copy
+from gatenlp.span import Span
 from gatenlp.annotation import Annotation
 from gatenlp.impl import SortedIntvls
-from gatenlp.utils import support_annotation_or_set
-from collections.abc import Iterable
+from gatenlp.utils import support_annotation_or_set, allowspan
 
 __pdoc__ = {
     "AnnotationSet.__iter__": True,
@@ -20,11 +23,12 @@ __pdoc__ = {
 
 class InvalidOffsetError(KeyError):
     """ """
+
     pass
 
 
 class AnnotationSet:
-    def __init__(self, name: str = "", owner_doc: "Document" = None):
+    def __init__(self, name: str = "", owner_doc=None):
         """
         Creates an annotation set. This should not be used directly by the user, instead the
         method `Document.annset(name)` should be used to access the annotation set with a given
@@ -44,6 +48,7 @@ class AnnotationSet:
         self._name = name
         self._owner_doc = owner_doc
         self._index_by_offset = None
+        self._index_by_ol = None
         self._index_by_type = None
         # internally we represent the annotations as a map from annotation id (int) to Annotation
         self._annotations = {}
@@ -77,11 +82,13 @@ class AnnotationSet:
             if self.__dict__.get(key, None) is None:
                 super().__setattr__(key, value)
             else:
-                raise Exception("AnnotationSet attribute cannot get changed after being set")
+                raise Exception(
+                    "AnnotationSet attribute cannot get changed after being set"
+                )
         else:
             super().__setattr__(key, value)
 
-    def detach(self, restrict_to=None) -> "AnnotationSet":
+    def detach(self, restrict_to=None):
         """
         Creates an immutable and detached copy of this set, optionally restricted to the given annotation ids.
         A detached annotation set does not have an owning document and deleting or adding annotations does not
@@ -95,21 +102,25 @@ class AnnotationSet:
         Returns:
           an immutable annotation set
         """
-        annset = AnnotationSet(name="detached-from:"+self.name)
+        annset = AnnotationSet(name="detached-from:" + self.name)
         annset._is_immutable = True
         if restrict_to is None:
-            annset._annotations = {annid: self._annotations[annid] for annid in self._annotations.keys()}
+            annset._annotations = {
+                annid: self._annotations[annid] for annid in self._annotations.keys()
+            }
         else:
-            annset._annotations = {annid: self._annotations[annid] for annid in restrict_to}
+            annset._annotations = {
+                annid: self._annotations[annid] for annid in restrict_to
+            }
         annset._next_annid = self._next_annid
         return annset
 
-    def detach_from(self, anns: Iterable) -> "AnnotationSet":
+    def detach_from(self, anns: Iterable):
         """
         Creates an immutable detached annotation set from the annotations in anns which could by
         either a collection of annotations or annotation ids (int numbers) which are assumed to
         be the annotation ids from this set.
-        
+
         The next annotation id for the created set is the highest seen annotation id from anns plus one.
 
         Args:
@@ -118,7 +129,7 @@ class AnnotationSet:
         Returns:
           an immutable detached annotation set
         """
-        annset = AnnotationSet(name="detached-from:"+self.name)
+        annset = AnnotationSet(name="detached-from:" + self.name)
         annset._is_immutable = True
         annset._annotations = {}
         nextid = -1
@@ -164,6 +175,15 @@ class AnnotationSet:
             for ann in self._annotations.values():
                 self._index_by_offset.add(ann.start, ann.end, ann.id)
 
+    def _create_index_by_ol(self) -> None:
+        """
+        Generates an index by start offset, end offset and annotation id
+        """
+        if self._index_by_ol is None:
+            self._index_by_ol = SortedIntvls(by_ol=True)
+            for ann in self._annotations.values():
+                self._index_by_ol.add(ann.start, ann.end, ann.id)
+
     def _create_index_by_type(self) -> None:
         """Generates the type index, if it does not already exist. The type index is a map from
         annotation type to a set of all annotation ids with that type.
@@ -179,7 +199,7 @@ class AnnotationSet:
 
         Args:
           annotation: the annotation to add to the indices.
-          annotation: Annotation: 
+          annotation: Annotation:
         """
         if self._index_by_type is not None:
             self._index_by_type[annotation.type].add(annotation.id)
@@ -191,10 +211,12 @@ class AnnotationSet:
 
         Args:
           annotation: the annotation to remove.
-          annotation: Annotation: 
+          annotation: Annotation:
         """
         if self._index_by_offset is not None:
-            self._index_by_offset.remove(annotation.start, annotation.end, annotation.id)
+            self._index_by_offset.remove(
+                annotation.start, annotation.end, annotation.id
+            )
         if self._index_by_type is not None:
             self._index_by_type[annotation.type].remove(annotation.id)
 
@@ -235,17 +257,19 @@ class AnnotationSet:
                 ret.add(i[2])
         return ret
 
-    def _restrict_intvs(self, intvs, ignore=None) -> "AnnotationSet":
+    def _restrict_intvs(self, intvs, ignore=None):
         """
 
         Args:
-          intvs: 
+          intvs:
           ignore:  (Default value = None)
 
         Returns:
 
         """
-        return self.detach(restrict_to=AnnotationSet._intvs2idlist(intvs, ignore=ignore))
+        return self.detach(
+            restrict_to=AnnotationSet._intvs2idlist(intvs, ignore=ignore)
+        )
 
     def __len__(self) -> int:
         """
@@ -263,7 +287,7 @@ class AnnotationSet:
         return len(self._annotations)
 
     @property
-    def document(self) -> Union["Document", None]:
+    def document(self):
         """
         Returns the owning document, if set. If the owning document was not set, returns None.
         """
@@ -276,8 +300,8 @@ class AnnotationSet:
         document and if the owning document has text.
 
         Args:
-          start: int: 
-          end: int: 
+          start: int:
+          end: int:
           annid:  (Default value = None)
         """
         if self._owner_doc is None:
@@ -294,10 +318,16 @@ class AnnotationSet:
             raise InvalidOffsetError("Annotation ends before it starts")
         if start > doc_size:
             raise InvalidOffsetError(
-                "Annotation starts after document ends: start={}, docsize={}".format(start, doc_size))
+                "Annotation starts after document ends: start={}, docsize={}".format(
+                    start, doc_size
+                )
+            )
         if end > doc_size:
             raise InvalidOffsetError(
-                "Annotation ends after document ends: end={}, docsize={}".format(end, doc_size))
+                "Annotation ends after document ends: end={}, docsize={}".format(
+                    end, doc_size
+                )
+            )
 
     @property
     def start(self):
@@ -318,7 +348,7 @@ class AnnotationSet:
         """
         Returns the end offset of the annotation set, i.e. the biggest end offset of any annotation.
         This needs the index and creates it if necessary.
-        
+
         Throws:
           an exception if there are no annotations in the set.
         """
@@ -335,9 +365,17 @@ class AnnotationSet:
         Throws:
           an exception if there are no annotations in the set.
         """
-        return self.end() - self.start()
+        return self.end - self.start
 
-    def add(self, start: int, end: int, anntype: str, features: Dict[str, Any] = None, annid: int = None):
+    @allowspan
+    def add(
+        self,
+        start: int,
+        end: int,
+        anntype: str,
+        features: Dict[str, Any] = None,
+        annid: int = None,
+    ):
         """
         Adds an annotation to the set. Once an annotation has been added, the start and end offsets,
         the type, and the annotation id of the annotation are immutable.
@@ -358,12 +396,16 @@ class AnnotationSet:
         if annid is not None and not isinstance(annid, int):
             raise Exception("Parameter annid must be an int, mixed up with features?")
         if features is not None and isinstance(features, int):
-            raise Exception("Parameter features must not be an int: mixed up with annid?")
+            raise Exception(
+                "Parameter features must not be an int: mixed up with annid?"
+            )
         if self._is_immutable:
             raise Exception("Cannot add an annotation to an immutable annotation set")
         self._check_offsets(start, end)
         if annid and annid in self._annotations:
-            raise Exception("Cannot add annotation with id {}, already in set".format(annid))
+            raise Exception(
+                "Cannot add annotation with id {}, already in set".format(annid)
+            )
         if annid is None:
             annid = self._next_annid
             self._next_annid = self._next_annid + 1
@@ -375,14 +417,14 @@ class AnnotationSet:
         self._add_to_indices(ann)
         if self.changelog is not None:
             entry = {
-                    "command": "annotation:add",
-                    "set": self.name,
-                    "start": ann.start,
-                    "end": ann.end,
-                    "type": ann.type,
-                    "features": ann._features.to_dict(),
-                    "id": ann.id
-                }
+                "command": "annotation:add",
+                "set": self.name,
+                "start": ann.start,
+                "end": ann.end,
+                "type": ann.type,
+                "features": ann._features.to_dict(),
+                "id": ann.id,
+            }
             self.changelog.append(entry)
         return ann
 
@@ -401,39 +443,58 @@ class AnnotationSet:
         """
         return self.add(ann.start, ann.end, ann.type, ann.features, annid=annid)
 
-    def remove(self, annotation: Union[int, Annotation]) -> None:
+    def remove(
+        self, annoriter: Union[int, Annotation, Iterable], raise_on_notexisting=True
+    ) -> None:
         """
-        Removes the given annotation which is either the id or the annotation instance.
+        Removes the given annotation which is either the id or the annotation instance or
+        recursively all annotations in the iterable.
 
         Throws:
             exception if the annotation set is immutable or the annotation is not in the set
 
         Args:
-          annotation: either the id (int) or the annotation instance (Annotation)
+          annoriter: either the id (int) or the annotation instance (Annotation) or an iterable of
+            id or annotation instance or iterable ...
+          raise_on_notexisting: (default: True) if false, silently accepts non-existing annotations/ids and does nothing.
+            Note: if this is True, but the annotation set is immutable, an Exception is still raised.
         """
-        annid = None  # make pycharm happy
         if self._is_immutable:
-            raise Exception("Cannot remove an annotation from an immutable annotation set")
-        if isinstance(annotation, int):
-            annid = annotation
+            raise Exception(
+                "Cannot remove an annotation from an immutable annotation set"
+            )
+        if isinstance(annoriter, Iterable):
+            for a in annoriter:
+                self.remove(a, raise_on_notexisting=raise_on_notexisting)
+            return
+        annid = None  # make pycharm happy
+        if isinstance(annoriter, int):
+            annid = annoriter
             if annid not in self._annotations:
-                raise Exception("Annotation with id {} not in annotation set, cannot remove".format(annid))
-            annotation = self._annotations[annid]
-        elif isinstance(annotation, Annotation):
-            annid = annotation.id
+                raise Exception(
+                    "Annotation with id {} not in annotation set, cannot remove".format(
+                        annid
+                    )
+                )
+            annoriter = self._annotations[annid]
+        elif isinstance(annoriter, Annotation):
+            annid = annoriter.id
             if annid not in self._annotations:
-                raise Exception("Annotation with id {} does not belong to this set, cannot remove".format(annid))
+                raise Exception(
+                    "Annotation with id {} does not belong to this set, cannot remove".format(
+                        annid
+                    )
+                )
         # NOTE: once the annotation has been removed from the set, it could still be referenced
         # somewhere else and its features could get modified. In order to prevent logging of such changes,
         # the owning set gets cleared for the annotation
-        annotation._owner_set = None
+        annoriter._owner_set = None
         del self._annotations[annid]
         if self.changelog is not None:
-            self.changelog.append({
-                "command": "annotation:remove",
-                "set": self.name,
-                "id": annid})
-        self._remove_from_indices(annotation)
+            self.changelog.append(
+                {"command": "annotation:remove", "set": self.name, "id": annid}
+            )
+        self._remove_from_indices(annoriter)
 
     def clear(self) -> None:
         """
@@ -443,9 +504,7 @@ class AnnotationSet:
         self._index_by_offset = None
         self._index_by_type = None
         if self.changelog is not None:
-            self.changelog.append({
-                "command": "annotations:clear",
-                "set": self.name})
+            self.changelog.append({"command": "annotations:clear", "set": self.name})
 
     def clone_anns(self, memo=None):
         """
@@ -515,13 +574,67 @@ class AnnotationSet:
             for annid, ann in self._annotations.items():
                 yield ann
 
-    def iter(self,
-             start_ge: Union[int, None] = None,
-             start_lt: Union[None, int] = None,
-             with_type: str = None,
-             reverse: bool = False) -> Generator:
+    def iter(
+        self,
+        start_ge: Union[int, None] = None,
+        start_lt: Union[None, int] = None,
+        with_type: str = None,
+        reverse: bool = False,
+    ) -> Generator:
         """
-        Yields annotations in document order, otionally limited
+        Yields annotations ordered by starting annotation and annotation id, otionally limited
+        by the other parameters.
+
+        Args:
+          start_ge: the offset from where to start including annotations
+          start_lt: the last offset to use as the starting offset of an annotation
+          with_type: only annotations of this type
+          reverse: process in reverse document order
+
+        Yields:
+          annotations in document order
+
+        """
+
+        if with_type is not None:
+            allowedtypes = set()
+            if isinstance(type, str):
+                allowedtypes.add(with_type)
+            else:
+                for atype in with_type:
+                    allowedtypes.add(atype)
+        else:
+            allowedtypes = None
+        if not self._annotations:
+            return
+        maxoff = None
+        if start_ge is not None:
+            assert start_ge >= 0
+        if start_lt is not None:
+            assert start_lt >= 1
+            maxoff = start_lt + 1
+        if start_lt is not None and start_ge is not None:
+            assert start_lt > start_ge
+        self._create_index_by_offset()
+        for _start, _end, annid in self._index_by_offset.irange(
+            minoff=start_ge, maxoff=maxoff, reverse=reverse
+        ):
+            if (
+                allowedtypes is not None
+                and self._annotations[annid].type not in allowedtypes
+            ):
+                continue
+            yield self._annotations[annid]
+
+    def iter_ol(
+        self,
+        start_ge: Union[int, None] = None,
+        start_lt: Union[None, int] = None,
+        with_type: str = None,
+        reverse: bool = False,
+    ) -> Generator:
+        """
+        Yields annotations ordered by start offset, end offset and annotoation id, otionally limited
         by the other parameters. If two annoations start at the same offset, they are always
         ordered by increasing annotation id.
 
@@ -555,9 +668,14 @@ class AnnotationSet:
             maxoff = start_lt + 1
         if start_lt is not None and start_ge is not None:
             assert start_lt > start_ge
-        self._create_index_by_offset()
-        for _start, _end, annid in self._index_by_offset.irange(minoff=start_ge, maxoff=maxoff, reverse=reverse):
-            if allowedtypes is not None and self._annotations[annid].type not in allowedtypes:
+        self._create_index_by_ol()
+        for _start, _end, annid in self._index_by_ol.irange(
+            minoff=start_ge, maxoff=maxoff, reverse=reverse
+        ):
+            if (
+                allowedtypes is not None
+                and self._annotations[annid].type not in allowedtypes
+            ):
                 continue
             yield self._annotations[annid]
 
@@ -575,16 +693,18 @@ class AnnotationSet:
         """
         return self.iter(reverse=True, **kwargs)
 
-    def get(self, annid: Union[int, Annotation], default=None) -> Union[Annotation, None]:
+    def get(
+        self, annid: Union[int, Annotation], default=None
+    ) -> Union[Annotation, None]:
         """Gets the annotation with the given annotation id or returns the given default.
-        
+
         NOTE: for handling cases where legacy code still expects the add method to return
         an id and not the annotation, this will accept an annotation so the the frequent
         pattern still works:
-        
+
            annid = annset.add(b,e,t).id
            ann = annset.get(annid)
-        
+
         If an annotation is passed the annotation from the set with the id of that annotation is
         returned, if the annotation is from that set, this will return the same object, if it is
         still in the set (or return the default value).
@@ -592,8 +712,8 @@ class AnnotationSet:
         Args:
           annid: the annotation id of the annotation to retrieve.
           default: what to return if an annotation with the given id is not found. (Default value = None)
-          annid: Union[int: 
-          Annotation]: 
+          annid: Union[int:
+          Annotation]:
 
         Returns:
           the annotation or the default value.
@@ -605,11 +725,10 @@ class AnnotationSet:
 
     def first(self):
         """
-
-        Args:
+        Return the first (or only) annotation in the set by offset.
 
         Returns:
-          :return: first annotation
+          first annotation
 
         """
         sz = len(self._annotations)
@@ -623,11 +742,10 @@ class AnnotationSet:
 
     def last(self):
         """
-
-        Args:
+        Return the last (or only) annotation by offset.
 
         Returns:
-          :return: first annotation
+          last annotation
 
         """
         sz = len(self._annotations)
@@ -639,6 +757,26 @@ class AnnotationSet:
         _, _, annid = next(self._index_by_offset.irange(reverse=True))
         return self._annotations[annid]
 
+    def for_idx(self, idx, default=None):
+        """
+        Return the annotation corresponding to the index idx in the set. This returns the
+        annotation stored at the index, as added to the set. The order usually depends on the insertion time.
+        If no annotation with the given index is specified, the value specified for `default` is returned.
+
+        Args:
+            idx:  index of the annotation in the set
+            default: default value to return if now annotation with the given index exists
+
+        Returns:
+            the annotation with the given index or the default value
+        """
+        # TODO: we could make this more memory efficient (but slower) by iterating over values until getting idxth
+        tmplist = list(self._annotations.values())
+        if idx < len(tmplist):
+            return tmplist[idx]
+        else:
+            return default
+
     def __getitem__(self, annid):
         """
         Gets the annotation with the given annotation id or throws an exception.
@@ -651,8 +789,7 @@ class AnnotationSet:
         """
         return self._annotations[annid]
 
-    def with_type(self, *anntype: Union[str, Iterable],
-                  non_overlapping: bool = False) -> "AnnotationSet":
+    def with_type(self, *anntype: Union[str, Iterable], non_overlapping: bool = False):
         """
         Gets annotations of the specified type(s).
         Creates the type index if necessary.
@@ -709,7 +846,7 @@ class AnnotationSet:
             # then skip to the next group that does not overlap with the one we just selected
             typepriority = dict()
             for i, atype in enumerate(atypes):
-                typepriority[atype] = len(atypes)-i
+                typepriority[atype] = len(atypes) - i
             curminoffset = 0
             for group in allannsgrouped:
                 # instead of sorting, go through the group and find the top priority one
@@ -724,7 +861,7 @@ class AnnotationSet:
                         if ann.start >= curminoffset:
                             topann = ann
                             break
-                    for ann in group[i+1:]:
+                    for ann in group[i + 1:]:
                         if ann.start < curminoffset:
                             continue
                         if typepriority[ann.type] > typepriority[topann.type]:
@@ -761,13 +898,13 @@ class AnnotationSet:
 
     def by_span(self):
         """
-        Yields list of annotations with identical spans.
+        Yields list of annotations with identical spans. Note: first needs to sort all annotations!
         """
         self._create_index_by_offset()
         lastsoff = -1
         lasteoff = -1
         curlist = []
-        for ann in self.iter():
+        for ann in self.iter_ol():
             if ann.start != lastsoff or ann.end != lasteoff:
                 if lastsoff != -1:
                     yield curlist
@@ -788,7 +925,9 @@ class AnnotationSet:
         return self._index_by_type.keys()
 
     @support_annotation_or_set
-    def start_eq(self, start: int, ignored: Any = None, annid=None, include_self=False) -> "AnnotationSet":
+    def startingat(
+        self, start: int, ignored: Any = None, annid=None, include_self=False
+    ):
         """
         Gets all annotations starting at the given offset (empty if none) and returns them in a detached
         annotation set.
@@ -806,7 +945,7 @@ class AnnotationSet:
             detached annotation set of matching annotations
         """
         self._create_index_by_offset()
-        intvs = self._index_by_offset.starting_from(start)
+        intvs = self._index_by_offset.starting_at(start)
         if not include_self and annid is not None:
             ignore = annid
         else:
@@ -814,7 +953,9 @@ class AnnotationSet:
         return self._restrict_intvs(intvs, ignore=ignore)
 
     @support_annotation_or_set
-    def start_min_ge(self, offset: int, ignored: Any = None, annid=None, include_self=False) -> "AnnotationSet":
+    def start_min_ge(
+        self, offset: int, ignored: Any = None, annid=None, include_self=False
+    ):
         """Gets all annotations starting at the first possible offset
         at or after the given offset and returns them in an immutable
         annotation set.
@@ -857,7 +998,7 @@ class AnnotationSet:
         return self.detach(restrict_to=retids)
 
     @support_annotation_or_set
-    def start_ge(self, start: int, ignored: Any = None, annid=None, include_self=False) -> "AnnotationSet":
+    def start_ge(self, start: int, ignored: Any = None, annid=None, include_self=False):
         """Return the annotations that start at or after the given start offset.
 
         Args:
@@ -879,7 +1020,7 @@ class AnnotationSet:
         return self._restrict_intvs(intvs, ignore=ignore)
 
     @support_annotation_or_set
-    def start_lt(self, offset: int, ignored: Any = None, annid=None) -> "AnnotationSet":
+    def start_lt(self, offset: int, ignored: Any = None, annid=None):
         """
         Returns the annotations that start before the given offset (or annotation). This also accepts an annotation
         or set.
@@ -897,13 +1038,12 @@ class AnnotationSet:
         intvs = self._index_by_offset.starting_before(offset)
         return self._restrict_intvs(intvs)
 
-
     @support_annotation_or_set
-    def overlapping(self, start: int, end: int, annid=None, include_self=False) -> "AnnotationSet":
+    def overlapping(self, start: int, end: int, annid=None, include_self=False):
         """
         Gets annotations overlapping with the given span. Instead of the start and end offsets,
         also accepts an annotation or annotation set.
-        
+
         For each annotation ann in the result set, ann.overlapping(span) is True
 
         Args:
@@ -926,11 +1066,11 @@ class AnnotationSet:
         return self._restrict_intvs(intvs, ignore=ignore)
 
     @support_annotation_or_set
-    def covering(self, start: int, end: int, annid=None, include_self=False) -> "AnnotationSet":
+    def covering(self, start: int, end: int, annid=None, include_self=False):
         """
         Gets the annotations which contain the given offset range (or annotation/annotation set),
         i.e. annotations such that the given offset range is within the annotation.
-        
+
         For each annotation ann in the result set, ann.covering(span) is True.
 
         Args:
@@ -953,11 +1093,11 @@ class AnnotationSet:
         return self._restrict_intvs(intvs, ignore=ignore)
 
     @support_annotation_or_set
-    def within(self, start: int, end: int, annid=None, include_self=False) -> "AnnotationSet":
+    def within(self, start: int, end: int, annid=None, include_self=False):
         """
         Gets annotations that fall completely within the given offset range, i.e. annotations
         such that the offset range is covering each of the annotation.
-        
+
         For each annotation ann in the result set, ann.within(span) is True.
 
         Args:
@@ -971,9 +1111,7 @@ class AnnotationSet:
           an immutable annotation set with the matching annotations
 
         """
-        if start == end:
-            intvs = []
-        elif start > end:
+        if start > end:
             raise Exception("Invalid offset range: {},{}".format(start, end))
         else:
             self._create_index_by_offset()
@@ -985,10 +1123,10 @@ class AnnotationSet:
         return self._restrict_intvs(intvs, ignore=ignore)
 
     @support_annotation_or_set
-    def coextensive(self, start: int, end: int, annid=None, include_self=False) -> "AnnotationSet":
+    def coextensive(self, start: int, end: int, annid=None, include_self=False):
         """
         Returns a detached annotation set with all annotations that start and end at the given offsets.
-        
+
         For each annotation ann in the result set, ann.coextensive(span) is True.
 
         Args:
@@ -1009,15 +1147,83 @@ class AnnotationSet:
             ignore = None
         return self._restrict_intvs(intvs, ignore=ignore)
 
+    @support_annotation_or_set
+    def before(
+        self, start: int, end: int, annid=None, include_self=False, immediately=False
+    ):
+        """
+        Returns a detached annotation set with all annotations that end before the given offsets.
+
+        For each annotation ann in the result set, ann.isbefore(span) is True.
+
+        Args:
+            start: start offset of the span
+            end: end offset of the span
+            annid: the annotation id of the annotation representing the span. (Default value = None)
+            include_self: if True and the annotation id for the span is given, do not include that
+                annotation in the result set.
+            immediately: if True, the end offset of the annotations return must coincide with the
+                start offset of the span (default=False)
+
+        Returns:
+          annotation set with all annotations that end before the given span
+        """
+        self._create_index_by_offset()
+        if immediately:
+            intvs = self._index_by_offset.ending_at(start)
+        else:
+            intvs = self._index_by_offset.ending_to(start)
+        # we need to filter self if self is zero-length!
+        if not include_self and annid is not None:
+            ignore = annid
+        else:
+            ignore = None
+        return self._restrict_intvs(intvs, ignore=ignore)
+
+    @support_annotation_or_set
+    def after(
+        self, start: int, end: int, annid=None, include_self=False, immediately=False
+    ):
+        """
+        Returns a detached annotation set with all annotations that start after the given span.
+
+        For each annotation ann in the result set, ann.isafter(span) is True.
+
+        Args:
+            start: start offset of the span
+            end: end offset of the span
+            annid: the annotation id of the annotation representing the span. (Default value = None)
+            include_self: if True and the annotation id for the span is given, do not include that
+                annotation in the result set.
+            immediately: if True, the start offset of the annotations returned must coincide with the
+                end offset of the span (default=False)
+
+        Returns:
+          annotation set with all annotations that start after the given span
+        """
+        self._create_index_by_offset()
+        if immediately:
+            intvs = self._index_by_offset.starting_at(end)
+        else:
+            intvs = self._index_by_offset.starting_from(end)
+        # we need to filter self if self is zero-length!
+        if not include_self and annid is not None:
+            ignore = annid
+        else:
+            ignore = None
+        return self._restrict_intvs(intvs, ignore=ignore)
+
     @property
-    def span(self) -> Tuple[int, int]:
+    def span(self) -> Span:
         """
         Returns a tuple with the start and end offset the corresponds to the smallest start offset of any annotation
         and the largest end offset of any annotation.
         (Builds the offset index)
         """
+        if len(self._annotations) == 0:
+            return Span(0, 0)
         self._create_index_by_offset()
-        return self._index_by_offset.min_start(), self._index_by_offset.max_end()
+        return Span(self._index_by_offset.min_start(), self._index_by_offset.max_end())
 
     def __contains__(self, annorannid: Union[int, Annotation]) -> bool:
         """
@@ -1031,7 +1237,9 @@ class AnnotationSet:
         """
         if isinstance(annorannid, Annotation):
             return annorannid.id in self._annotations
-        return annorannid in self._annotations  # On the off chance someone passed an ID in directly
+        return (
+            annorannid in self._annotations
+        )  # On the off chance someone passed an ID in directly
 
     contains = __contains__
 
@@ -1041,21 +1249,32 @@ class AnnotationSet:
         """
         return "AnnotationSet({})".format(repr(list(self.iter())))
 
-    def to_dict(self, **kwargs):
+    def to_dict(self, anntypes=None, **kwargs):
         """
         Convert an annotation set to its dict representation.
 
         Args:
-          **kwargs: passed on to the dict creation of contained annotations.
+            anntypes: if not None, an iterable of annotation types to include
+            **kwargs: passed on to the dict creation of contained annotations.
 
         Returns:
             the dict representation of the annotation set.
         """
+        if anntypes is not None:
+            anntypesset = set(anntypes)
+            anns_list = list(
+                val.to_dict(**kwargs)
+                for val in self._annotations.values()
+                if val.type in anntypesset
+            )
+        else:
+            anns_list = list(
+                val.to_dict(**kwargs) for val in self._annotations.values()
+            )
         return {
             # NOTE: Changelog is not getting added as it is stored in the document part!
             "name": self.name,
-            "annotations": list(val.to_dict(**kwargs)
-                                for val in self._annotations.values()),
+            "annotations": anns_list,
             "next_annid": self._next_annid,
         }
 
@@ -1077,9 +1296,36 @@ class AnnotationSet:
         if dictrepr.get("annotations"):
             annset._annotations = dict(
                 (int(a["id"]), Annotation.from_dict(a, owner_set=annset, **kwargs))
-                for a in dictrepr.get("annotations"))
+                for a in dictrepr.get("annotations")
+            )
         else:
             annset._annotations = {}
         return annset
 
+    @staticmethod
+    def from_anns(anns, deep_copy=False, **kwargs):
+        """
+        Create a detached AnnotationSet from an iterable of annotations.
 
+        Args:
+          anns: an iterable of annotations
+          deep_copy: if the annotations should get added as copies (default) or deep copies.
+
+        Returns:
+            the annotation set
+        """
+        annset = AnnotationSet(name="", owner_doc=None)
+        annset._annotations = dict()
+        maxid = 0
+        for ann in anns:
+            if deep_copy:
+                addann = ann.deepcopy()
+            else:
+                addann = ann.copy()
+            annset._annotations[addann.id] = addann
+            if addann.id > maxid:
+                maxid = addann.id
+        annset._next_annid = maxid
+        annset._is_immutable = True
+
+        return annset
