@@ -17,21 +17,24 @@ from gatenlp.offsetmapper import OffsetMapper
 # * send raw HTML or other formats support by the endpoint instead "doc" (which so far is just text)
 # * maybe support the 100-continue protocol so far we dont
 # * ERROR HANDLING: raise exception vs return None?
+
+
 class GateCloudAnnotator(Annotator):
     """
     This annotator sends the text of a document to a GATE Cloud (https://cloud.gate.ac.uk/) endpoint and uses the
     returned result to create annotations.
     """
 
-    def __init__(self,
-                 api_key=None,
-                 api_password=None,
-                 url=None,
-                 ann_types=None,
-                 map_types=None,
-                 out_annset="",
-                 min_delay_ms=501,
-                 ):
+    def __init__(
+        self,
+        api_key=None,
+        api_password=None,
+        url=None,
+        ann_types=None,
+        map_types=None,
+        out_annset="",
+        min_delay_ms=501,
+    ):
         """
         Create a GateCloudAnnotator.
 
@@ -60,9 +63,13 @@ class GateCloudAnnotator(Annotator):
             if isinstance(ann_types, str):
                 self.ann_types = ann_types
             elif isinstance(ann_types, list):
-                self.ann_types = ",".join([at if at.startswith(":") else ":"+at for at in ann_types])
+                self.ann_types = ",".join(
+                    [at if at.startswith(":") else ":" + at for at in ann_types]
+                )
             else:
-                raise Exception("ann_types mist be a string of types like ':Person,:Location' or a list of types")
+                raise Exception(
+                    "ann_types mist be a string of types like ':Person,:Location' or a list of types"
+                )
         else:
             self.ann_types = None
         self.logger = init_logger()
@@ -78,19 +85,29 @@ class GateCloudAnnotator(Annotator):
         else:
             url = self.url
         text = doc.text
-        hdrs = {'Content-Type': 'text/plain; charset=UTF-8', 'Accept': 'application/gate+json'}
+        hdrs = {
+            "Content-Type": "text/plain; charset=UTF-8",
+            "Accept": "application/gate+json",
+        }
         params = {}
         if self.ann_types:
             params["annotations"] = self.ann_types
         # NOTE: not sure when this is needed, for now, disabled
-        #next_annid = doc.annset(self.out_annset)._next_annid
-        #params["nextAnnotationId"] = str(next_annid)
+        # next_annid = doc.annset(self.out_annset)._next_annid
+        # params["nextAnnotationId"] = str(next_annid)
         # self.logger.debug(f"Sending text={text}, params={params}")
         if self.api_key:
-            response = requests.post(url, data=text.encode("utf-8"), headers=hdrs, params=params,
-                                     auth=HTTPBasicAuth(self.api_key, self.api_password))
+            response = requests.post(
+                url,
+                data=text.encode("utf-8"),
+                headers=hdrs,
+                params=params,
+                auth=HTTPBasicAuth(self.api_key, self.api_password),
+            )
         else:
-            response = requests.post(url, data=text.encode("utf-8"), headers=hdrs, params=params)
+            response = requests.post(
+                url, data=text.encode("utf-8"), headers=hdrs, params=params
+            )
         scode = response.status_code
         if scode != 200:
             raise Exception(f"Something went wrong, received status code {scode}")
@@ -100,7 +117,10 @@ class GateCloudAnnotator(Annotator):
         for typename, anns in ents.items():
             for anndata in anns:
                 feats = {}
-                start, end = None, None   # cause an exception if the return data does not have indices
+                start, end = (
+                    None,
+                    None,
+                )  # cause an exception if the return data does not have indices
                 for fname, fval in anndata.items():
                     if fname == "indices":
                         start, end = fval[0], fval[1]
@@ -118,20 +138,22 @@ class TagMeAnnotator(Annotator):
     An annotator that sends text to the TagMe Annotation service (https://sobigdata.d4science.org/group/tagme/tagme)
     and uses the result to annotate the document.
     """
-    def __init__(self,
-                 lang="en",
-                 ann_type="Mention",
-                 auth_token=None,
-                 url=None,
-                 task="tag", # or spot
-                 out_annset="",
-                 min_delay_ms=501,
-                 tweet=False,
-                 include_all_spots=False,
-                 long_text=None,
-                 epsilon=None,
-                 link_pattern="https://{0}.wikipedia.org/wiki/{1}"
-                 ):
+
+    def __init__(
+        self,
+        url=None,
+        auth_token=None,
+        lang="en",
+        ann_type="Mention",
+        task="tag",  # or spot
+        out_annset="",
+        min_delay_ms=501,
+        tweet=False,
+        include_all_spots=False,
+        long_text=None,
+        epsilon=None,
+        link_pattern="https://{0}.wikipedia.org/wiki/{1}",
+    ):
         """
         Create a TagMeAnnotator.
 
@@ -158,7 +180,7 @@ class TagMeAnnotator(Annotator):
                 url = "https://tagme.d4science.org/tagme/spot"
             else:
                 raise Exception("task must be 'tag' or 'spot'")
-        assert lang in ['en', 'de', 'it']
+        assert lang in ["en", "de", "it"]
         if long_text is not None:
             assert isinstance(long_text, int)
         if epsilon is not None:
@@ -179,7 +201,7 @@ class TagMeAnnotator(Annotator):
         self.link_pattern = link_pattern
 
     def __call__(self, doc, **kwargs):
-        if 'tweet' in kwargs:
+        if "tweet" in kwargs:
             tweet = kwargs["tweet"]
         else:
             tweet = self.tweet
@@ -187,9 +209,14 @@ class TagMeAnnotator(Annotator):
         if delay < self.min_delay_s:
             time.sleep(self.min_delay_s - delay)
         text = doc.text
-        hdrs = {'Content-Type': 'text/plain; charset=UTF-8', 'Accept': 'application/gate+json'}
+        hdrs = {
+            "Content-Type": "text/plain; charset=UTF-8",
+            "Accept": "application/gate+json",
+        }
         params = {
-            'text': text, 'gcube-token': self.auth_token, 'lang': self.lang,
+            "text": text,
+            "gcube-token": self.auth_token,
+            "lang": self.lang,
         }
         if self.include_all_spots:
             params["include_all_spots"] = "true"
@@ -235,14 +262,16 @@ class TextRazorTextAnnotator(Annotator):
 
     NOTE: this annotator and how it can get parametrized will still change!
     """
-    def __init__(self,
-                 lang=None,  # if None/not specified, TextRazor auto-detects
-                 auth_token=None,
-                 url=None,  # use default
-                 extractors=None,
-                 out_annset="",
-                 min_delay_ms=501,
-                 ):
+
+    def __init__(
+        self,
+        url=None,  # use default
+        auth_token=None,
+        lang=None,  # if None/not specified, TextRazor auto-detects
+        extractors=None,
+        out_annset="",
+        min_delay_ms=501,
+    ):
         """
         Create a TextRazorTextAnnotator.
 
@@ -288,11 +317,9 @@ class TextRazorTextAnnotator(Annotator):
             # 'Content-Type': 'text/plain; charset=UTF-8',
             # 'Accept-encoding': 'gzip'  # TODO: to enable compressed responses
             # 'Content-encoding': 'gzip'  # TODO: to enable compressed requests
-            'X-TextRazor-Key': self.auth_token
-        }
-        data = {
-            'text': text.encode("UTF-8")
+            "X-TextRazor-Key": self.auth_token
         }
+        data = {"text": text.encode("UTF-8")}
         if self.extractors:
             data["extractors"] = self.extractors
         if self.lang:
@@ -302,7 +329,8 @@ class TextRazorTextAnnotator(Annotator):
             self.url,
             # params=params,
             data=data,
-            headers=hdrs)
+            headers=hdrs,
+        )
         scode = response.status_code
         if scode != 200:
             raise Exception(f"Something went wrong, received status code {scode}")
@@ -348,8 +376,17 @@ class TextRazorTextAnnotator(Annotator):
                 annset.add(sentstart, sentend, "Sentence")
         for ent in entities:
             feats = {}
-            for fname in ["wikiLink", "entityEnglishId", "wikidataId", "relevanceScore", "confidenceScore", "type",
-                          "freebaseId", "entityId", "freebaseTypes"]:
+            for fname in [
+                "wikiLink",
+                "entityEnglishId",
+                "wikidataId",
+                "relevanceScore",
+                "confidenceScore",
+                "type",
+                "freebaseId",
+                "entityId",
+                "freebaseTypes",
+            ]:
                 if fname in ent:
                     feats[fname] = ent[fname]
             annset.add(ent["startingPos"], ent["endingPos"], "Entity", feats)
@@ -357,33 +394,120 @@ class TextRazorTextAnnotator(Annotator):
 
 
 class ElgTextAnnotator(Annotator):
+    # TODO: maybe we should eventually always use the elg package and the elg Service class!
+    # TODO: however, currently their way how handling auth is done is too limiting see issues #8, #9
+
+    # TODO: use template and return the URL from a method or use elg.utils
+    ELG_SC_LIVE_URL_PREFIX = "https://live.european-language-grid.eu/auth/realms/ELG/protocol/openid-connect/auth?"
+    ELG_SC_LIVE_URL_PREFIX += (
+        "client_id=python-sdk&redirect_uri=urn:ietf:wg:oauth:2.0:oob&response_type=code"
+    )
+    ELG_SC_LIVE_URL_OFFLINE = ELG_SC_LIVE_URL_PREFIX + "&scope=offline_access"
+    ELG_SC_LIVE_URL_OPENID = ELG_SC_LIVE_URL_PREFIX + "&scope=openid"
+
+    ELG_SC_DEV_URL_PREFIX = "https://dev.european-language-grid.eu/auth/realms/ELG/protocol/openid-connect/auth?"
+    ELG_SC_DEV_URL_PREFIX += (
+        "client_id=python-sdk&redirect_uri=urn:ietf:wg:oauth:2.0:oob&response_type=code"
+    )
+    ELG_SC_DEV_URL_OFFLINE = ELG_SC_DEV_URL_PREFIX + "&scope=offline_access"
+    ELG_SC_DEV_URL_OPENID = ELG_SC_DEV_URL_PREFIX + "&scope=openid"
     """
     An annotator that sends text to one of the services registered with the European Language Grid
     (https://live.european-language-grid.eu/) and uses the result to create annotations.
 
     NOTE: This is maybe not properly implemented and not properly tested yet!
     """
-    def __init__(self,
-                 auth_token=None,
-                 url=None,
-                 out_annset="",
-                 min_delay_ms=501,
-                 anntypes_map=None,
-                 ):
+
+    def __init__(
+        self,
+        url=None,
+        service=None,
+        auth=None,
+        success_code=None,
+        access_token=None,
+        refresh_access=False,
+        out_annset="",
+        min_delay_ms=501,
+        anntypes_map=None,
+    ):
         """
         Create an ElgTextAnnotator.
 
+        NOTE: error handling is not properly implemented yet since we do not know yet how exactly the various
+        error conditions are represented in the result returned from the ELG services. For now, any error will
+        throw an exception when `__call__` is invoked.
+
+        NOTE: initialization can fail with an exception if success_code is specified and retrieving the
+        authentification information fails.
+
         Args:
-            auth_token: the authentication token from ELG
-            url:  the annotation service URL to use
+            url:  the annotation service URL to use. If not specified, the service parameter must be specified.
+            service: the ELG service number or a tuple (servicenumber, domain). This requires the elg package.
+                This may raise an exception. If successful, the url and service_meta attributes are set.
+            auth: a pre-initialized ELG Authentication object. Requires the elg package. If not specified, the
+                success_code or access_token parameter must be specified.
+            success_code: the success code returned from the ELG web page for one of the URLs to obtain
+                success codes. This will try to obtain the authentication information and store it in the
+                `auth` attribute.  Requires the elg package.
+                To obtain a success code, go the the ELG_SC_LIVE_URL_OPENID or ELG_SC_LIVE_URL_OFFLINE url
+                and log in with your ELG user id, this will show the success code that can be copy-pasted.
+            access_token: the access token token for the ELG service. Only used if auth or success_code are not
+                specified. The access token is probably only valid for a limited amount of time. No refresh
+                will be done and once the access token is invalid, calling `__call__` will fail with an exception.
+                The access token can be obtained using the elg package or copied from the "Code samples" tab
+                on the web page for a service after logging in.
+            refresh_access: if True, will try to refresh the access token if auth or success_code was specified and
+                refreshing is possible. Ignored if only access_token was specified
             out_annset: the name of the annotation set where to create the annotations (default: "")
             min_delay_ms: the minimum delay time between requests in milliseconds (default: 501 ms)
             anntypes_map: a map for renaming the annotation type names from the service to the ones to use in
                the annotated document.
         """
-        assert url is not None
-        self.auth_token = auth_token
+        if [x is not None for x in [url, service]].count(True) != 1:
+            raise Exception("Exactly one of service or url must be specified")
+        if [x is not None for x in [auth, success_code, access_token]].count(True) != 1:
+            raise Exception(
+                "Exactly one of auth, success_code, or access_token must be specified"
+            )
+        self.access_token = access_token
+        self.success_code = success_code
+        self.auth = auth
         self.url = url
+        self.service = service
+        self.service_meta = None
+        self.refresh_access = refresh_access
+        # first check if we need to import the elg package
+        import_elg = False
+        if access_token:
+            self.refresh_access = False
+        if service is not None:
+            import_elg = True
+        if auth or success_code:
+            import_elg = True
+        if import_elg:
+            try:
+                from elg import Authentication
+                from elg.utils import get_domain, get_metadatarecord
+            except Exception as ex:
+                raise Exception(
+                    "For this gatenlp must be installed with extra elg or extra all, e.g. gatenlp[elg]",
+                    ex,
+                )
+        if service is not None:
+            # update this to use the new method:
+            # https://gitlab.com/european-language-grid/platform/python-client/-/issues/9
+            if isinstance(service, tuple):
+                service_id, domain = service
+            else:
+                service_id = service
+                domain = get_domain("live")
+            self.service_meta = get_metadatarecord(service_id, domain)
+            # NOTE: there is also elg_execution_location for async requests!
+            self.url = self.service_meta["service_info"]["elg_execution_location_sync"]
+        if success_code is not None:
+            self.auth = Authentication.from_success_code(success_code, domain="live")
+        if self.auth:
+            self.access_token = self.auth.access_token
         self.min_delay_s = min_delay_ms / 1000.0
         self.anntypes_map = anntypes_map
         self.out_annset = out_annset
@@ -392,18 +516,25 @@ class ElgTextAnnotator(Annotator):
         self._last_call_time = 0
 
     def __call__(self, doc, **kwargs):
+        # if necessary and possible, refresh the access token
+        if self.refresh_access and self.auth:
+            self.auth.refresh_if_needed()
         delay = time.time() - self._last_call_time
         if delay < self.min_delay_s:
             time.sleep(self.min_delay_s - delay)
         om = OffsetMapper(doc.text)
-        request_json = json.dumps({"type": "text", "content": doc.text, "mimeType": "text/plain"})
-        hdrs = {'Content-Type': 'application/json'}
-        if self.auth_token:
-            hdrs["Authorization"] = f"Bearer {self.auth_token}"
+        request_json = json.dumps(
+            {"type": "text", "content": doc.text, "mimeType": "text/plain"}
+        )
+        hdrs = {"Content-Type": "application/json"}
+        if self.access_token:
+            hdrs["Authorization"] = f"Bearer {self.access_token}"
         response = requests.post(self.url, data=request_json, headers=hdrs)
         scode = response.status_code
         if scode != 200:
-            raise Exception(f"Something went wrong, received status code/text {scode} / {response.text}")
+            raise Exception(
+                f"Something went wrong, received status code/text {scode} / {response.text}"
+            )
         response_json = response.json()
         # self.logger.debug(f"Response JSON: {json}")
         # TODO: check that we have got
